@@ -16,7 +16,9 @@ import datetime
 
 import iso8601
 import mock
+import netaddr
 import six
+import testtools
 
 from oslo_versionedobjects import _utils
 from oslo_versionedobjects import base as obj_base
@@ -64,6 +66,32 @@ class FakeEnumField(fields.BaseEnumField):
     AUTO_TYPE = FakeEnum()
 
 
+class FakeStateMachineField(fields.StateMachine):
+
+    ACTIVE = 'ACTIVE'
+    PENDING = 'PENDING'
+    ERROR = 'ERROR'
+
+    ALLOWED_TRANSITIONS = {
+        ACTIVE: {
+            PENDING,
+            ERROR
+        },
+        PENDING: {
+            ACTIVE,
+            ERROR
+        },
+        ERROR: {
+            PENDING,
+        },
+    }
+
+    _TYPES = (ACTIVE, PENDING, ERROR)
+
+    def __init__(self, **kwargs):
+        super(FakeStateMachineField, self).__init__(self._TYPES, **kwargs)
+
+
 class FakeEnumAltField(fields.BaseEnumField):
     AUTO_TYPE = FakeEnumAlt()
 
@@ -119,6 +147,17 @@ class TestString(TestField):
         self.assertEqual("'123'", self.field.stringify(123))
 
 
+class TestSensitiveString(TestString):
+    def setUp(self):
+        super(TestSensitiveString, self).setUp()
+        self.field = fields.SensitiveStringField()
+
+    def test_stringify(self):
+        payload = """{'admin_password':'mypassword'}"""
+        expected = """'{'admin_password':'***'}'"""
+        self.assertEqual(expected, self.field.stringify(payload))
+
+
 class TestVersionPredicate(TestString):
     def setUp(self):
         super(TestVersionPredicate, self).setUp()
@@ -131,6 +170,64 @@ class TestVersionPredicate(TestString):
         self.from_primitive_values = self.coerce_good_values[0:1]
 
 
+class TestMACAddress(TestField):
+    def setUp(self):
+        super(TestMACAddress, self).setUp()
+        self.field = fields.MACAddressField()
+        self.coerce_good_values = [
+            ('c6:df:11:a5:c8:5d', 'c6:df:11:a5:c8:5d'),
+            ('C6:DF:11:A5:C8:5D', 'c6:df:11:a5:c8:5d'),
+            ('c6:df:11:a5:c8:5d', 'c6:df:11:a5:c8:5d'),
+            ('C6:DF:11:A5:C8:5D', 'c6:df:11:a5:c8:5d'),
+        ]
+        self.coerce_bad_values = [
+            'C6:DF:11:A5:C8',  # Too short
+            'C6:DF:11:A5:C8:5D:D7',  # Too long
+            'C6:DF:11:A5:C8:KD',  # Bad octal
+            1123123,  # Number
+            {},  # dict
+        ]
+        self.to_primitive_values = self.coerce_good_values[0:1]
+        self.from_primitive_values = self.coerce_good_values[0:1]
+
+
+class TestUUID(TestField):
+    def setUp(self):
+        super(TestUUID, self).setUp()
+
+    def test_validation_enabled(self):
+        self.field = fields.UUIDField()
+        self.coerce_good_values = [
+            ('da66a411-af0e-4829-9b67-475017ddd152',
+                'da66a411-af0e-4829-9b67-475017ddd152'),
+            ('da66a411af0e48299b67475017ddd152',
+                'da66a411af0e48299b67475017ddd152'),
+            ('DA66A411-AF0E-4829-9B67-475017DDD152',
+                'DA66A411-AF0E-4829-9B67-475017DDD152'),
+            ('DA66A411AF0E48299b67475017DDD152',
+                'DA66A411AF0E48299b67475017DDD152'),
+            # These values are included to ensure there is not change in
+            # behaviour - only when we can remove the old UUID behaviour can
+            #  we add these to the "self.coerce_bad_values" list
+            ('da66a411-af0e-4829-9b67',
+                'da66a411-af0e-4829-9b67'),
+            ('da66a411-af0e-4829-9b67-475017ddd152548999',
+                'da66a411-af0e-4829-9b67-475017ddd152548999'),
+            ('da66a411-af0e-4829-9b67-475017ddz152',
+                'da66a411-af0e-4829-9b67-475017ddz152'),
+            ('fake_uuid', 'fake_uuid'),
+            ('1', '1'),
+            (1, '1')
+        ]
+
+        self.to_primitive_values = self.coerce_good_values[0:1]
+        self.from_primitive_values = self.coerce_good_values[0:1]
+
+        self.test_coerce_good_values()
+        self.test_from_primitive()
+        self.test_to_primitive()
+
+
 class TestBaseEnum(TestField):
     def setUp(self):
         super(TestBaseEnum, self).setUp()
@@ -192,6 +289,71 @@ class TestEnum(TestField):
                           fields.EnumField, True)
 
 
+class TestStateMachine(TestField):
+
+    def test_good_transitions(self):
+        @obj_base.VersionedObjectRegistry.register
+        class AnObject(obj_base.VersionedObject):
+            fields = {
+                'status': FakeStateMachineField(),
+            }
+
+        obj = AnObject()
+
+        obj.status = FakeStateMachineField.ACTIVE
+        obj.status = FakeStateMachineField.PENDING
+        obj.status = FakeStateMachineField.ERROR
+        obj.status = FakeStateMachineField.PENDING
+        obj.status = FakeStateMachineField.ACTIVE
+
+    def test_bad_transitions(self):
+        @obj_base.VersionedObjectRegistry.register
+        class AnObject(obj_base.VersionedObject):
+            fields = {
+                'status': FakeStateMachineField(),
+            }
+
+        obj = AnObject(status='ERROR')
+
+        try:
+            obj.status = FakeStateMachineField.ACTIVE
+        except ValueError as e:
+            ex = e
+        else:
+            ex = None
+
+        self.assertIsNotNone(ex, 'Invalid transition failed to raise error')
+        self.assertEqual('AnObject.status is not allowed to transition out '
+                         'of \'ERROR\' state to \'ACTIVE\' state, choose from '
+                         '[\'PENDING\']',
+                         str(ex))
+
+    def test_bad_initial_value(self):
+        @obj_base.VersionedObjectRegistry.register
+        class AnObject(obj_base.VersionedObject):
+            fields = {
+                'status': FakeStateMachineField(),
+            }
+
+        obj = AnObject()
+
+        with testtools.ExpectedException(ValueError):
+            obj.status = "FOO"
+
+    def test_bad_updated_value(self):
+        @obj_base.VersionedObjectRegistry.register
+        class AnObject(obj_base.VersionedObject):
+            fields = {
+                'status': FakeStateMachineField(),
+            }
+
+        obj = AnObject()
+
+        with testtools.ExpectedException(ValueError):
+            obj.status = FakeStateMachineField.ACTIVE
+            obj.status = "FOO"
+
+
 class TestInteger(TestField):
     def setUp(self):
         super(TestField, self).setUp()
@@ -652,3 +814,87 @@ class TestObject(TestField):
         self.assertRaises(ValueError, pigs.coerce, None, "animal", ticktock)
         self.assertRaises(ValueError, pigs.coerce, None, "animal", wolfy)
         self.assertEqual(babe, pigs.coerce(None, "animal", babe))
+
+
+class TestIPAddress(TestField):
+    def setUp(self):
+        super(TestIPAddress, self).setUp()
+        self.field = fields.IPAddressField()
+        self.coerce_good_values = [('1.2.3.4', netaddr.IPAddress('1.2.3.4')),
+                                   ('::1', netaddr.IPAddress('::1')),
+                                   (netaddr.IPAddress('::1'),
+                                    netaddr.IPAddress('::1'))]
+        self.coerce_bad_values = ['1-2', 'foo']
+        self.to_primitive_values = [(netaddr.IPAddress('1.2.3.4'), '1.2.3.4'),
+                                    (netaddr.IPAddress('::1'), '::1')]
+        self.from_primitive_values = [('1.2.3.4',
+                                       netaddr.IPAddress('1.2.3.4')),
+                                      ('::1',
+                                       netaddr.IPAddress('::1'))]
+
+
+class TestIPAddressV4(TestField):
+    def setUp(self):
+        super(TestIPAddressV4, self).setUp()
+        self.field = fields.IPV4AddressField()
+        self.coerce_good_values = [('1.2.3.4', netaddr.IPAddress('1.2.3.4')),
+                                   (netaddr.IPAddress('1.2.3.4'),
+                                    netaddr.IPAddress('1.2.3.4'))]
+        self.coerce_bad_values = ['1-2', 'foo', '::1']
+        self.to_primitive_values = [(netaddr.IPAddress('1.2.3.4'), '1.2.3.4')]
+        self.from_primitive_values = [('1.2.3.4',
+                                       netaddr.IPAddress('1.2.3.4'))]
+
+
+class TestIPAddressV6(TestField):
+    def setUp(self):
+        super(TestIPAddressV6, self).setUp()
+        self.field = fields.IPV6AddressField()
+        self.coerce_good_values = [('::1', netaddr.IPAddress('::1')),
+                                   (netaddr.IPAddress('::1'),
+                                    netaddr.IPAddress('::1'))]
+        self.coerce_bad_values = ['1.2', 'foo', '1.2.3.4']
+        self.to_primitive_values = [(netaddr.IPAddress('::1'), '::1')]
+        self.from_primitive_values = [('::1',
+                                       netaddr.IPAddress('::1'))]
+
+
+class TestIPNetwork(TestField):
+    def setUp(self):
+        super(TestIPNetwork, self).setUp()
+        self.field = fields.IPNetworkField()
+        self.coerce_good_values = [('::1/0', netaddr.IPNetwork('::1/0')),
+                                   ('1.2.3.4/24',
+                                    netaddr.IPNetwork('1.2.3.4/24')),
+                                   (netaddr.IPNetwork('::1/32'),
+                                    netaddr.IPNetwork('::1/32'))]
+        self.coerce_bad_values = ['foo']
+        self.to_primitive_values = [(netaddr.IPNetwork('::1/0'), '::1/0')]
+        self.from_primitive_values = [('::1/0',
+                                       netaddr.IPNetwork('::1/0'))]
+
+
+class TestIPV4Network(TestField):
+    def setUp(self):
+        super(TestIPV4Network, self).setUp()
+        self.field = fields.IPV4NetworkField()
+        self.coerce_good_values = [('1.2.3.4/24',
+                                    netaddr.IPNetwork('1.2.3.4/24'))]
+        self.coerce_bad_values = ['foo', '::1/32']
+        self.to_primitive_values = [(netaddr.IPNetwork('1.2.3.4/24'),
+                                     '1.2.3.4/24')]
+        self.from_primitive_values = [('1.2.3.4/24',
+                                       netaddr.IPNetwork('1.2.3.4/24'))]
+
+
+class TestIPV6Network(TestField):
+    def setUp(self):
+        super(TestIPV6Network, self).setUp()
+        self.field = fields.IPV6NetworkField()
+        self.coerce_good_values = [('::1/0', netaddr.IPNetwork('::1/0')),
+                                   (netaddr.IPNetwork('::1/32'),
+                                    netaddr.IPNetwork('::1/32'))]
+        self.coerce_bad_values = ['foo', '1.2.3.4/24']
+        self.to_primitive_values = [(netaddr.IPNetwork('::1/0'), '::1/0')]
+        self.from_primitive_values = [('::1/0',
+                                       netaddr.IPNetwork('::1/0'))]
